#include "geometrie.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>


void couple_i_to_f( const Couple_i* const src, Couple_f* const dest ){
	dest->x = (float) src->x;
	dest->y = (float) src->y;
}

void couple_f_to_i( const Couple_f* const src, Couple_i* const dest ){
	dest->x = (int) src->x;
	dest->y = (int) src->y;
}

void array_of_couple_i_to_f( const Couple_i* src, Couple_f* dest, int n ){
	int i;
	for( i=0; i<n; i++ ){
		couple_i_to_f( src+i, dest+i );
	}
}

void array_of_couple_f_to_i( const Couple_f* src, Couple_i* dest, int n ){
	int i;
	for( i=0; i<n; i++ ){
		couple_f_to_i( src+i, dest+i );
	}
}

float distance( Point_f const * const A, Point_f const * const B ){
	return sqrt( (A->x - B->x)*(A->x - B->x) + (A->y - B->y)*(A->y - B->y) );
}

float norm_vector( const Point_f* u ){
	return sqrt( u->x*u->x + u->y*u->y );
}

int get_normalized_vector( const Vector_f* src, Vector_f* dest ){
	float n = norm_vector( src );
	if( n != 0 ){
		dest->x = src->x / n;
		dest->y = src->y / n;
	}else{
		dest->x = src->x;
		dest->y = src->y;
	}
	return n;
}

inline float scalar_product( const Vector_f* const u, const Vector_f* const v ){
	return u->x * v->x + u->y * v->y;
}

float vectorial_product( const Vector_f* u, const Vector_f* v ){
	return u->x * v->y - u->y * v->x;
}

float determinant( const Vector_f* u, const Vector_f* v ){
	return vectorial_product( u,v );
}

float cos_angle( const Vector_f* u, const Vector_f* v ){
	Vector_f normalized_u;
	get_normalized_vector( u, &normalized_u );
	Vector_f normalized_v;
	get_normalized_vector( v, &normalized_v );
	return scalar_product( &normalized_u, &normalized_v );
}

float sin_angle( const Vector_f* u, const Vector_f* v ){
	Vector_f normalized_u;
	get_normalized_vector( u, &normalized_u );
	Vector_f normalized_v;
	get_normalized_vector( v, &normalized_v );
	return vectorial_product( &normalized_u, &normalized_v );
}

float angle( const Vector_f* u, const Vector_f* v ){
	float angle = acos( cos_angle(u,v) );
	if( vectorial_product(u,v) >= 0 ){
		return angle;
	}else{
		return -angle;
	}
}

int intersection_point_and_vector( 
	const Point_f* A, const Vector_f* u,
	const Point_f* B, const Vector_f*  v,
	Point_f* intersection
){
	float inv_det = determinant( u,v );
	if( inv_det == 0 ) return 0;
	inv_det = 1/inv_det;
	Point_f tmp;
	tmp.x = vectorial_product( A, u );
	tmp.y = vectorial_product( B, v );
	intersection->x = inv_det * ( - (v->x * tmp.x) + (u->x * tmp.y) );
	intersection->y = inv_det * ( - (v->y * tmp.x) + (u->y * tmp.y) );
	return 1;
}

int intersection_point_and_point( 
	const Point_f* A1, const Point_f* A2,
	const Point_f* B1, const Point_f*  B2,
	Point_f* intersection
){
	Point_f u;
	u.x= A2->x - A1->x;
	u.y = A2->y - A1->y;
	Point_f v;
	v.x = B2->x - B1->x;
	v.y = B2->y - B1->y;
	return intersection_point_and_vector( A1, &u, B1, &v, intersection );
}

void center_of_fermat_circle( 
	const Point_f* A, const Point_f* B, Point_f* fermat_point
){
	fermat_point->x = 0.5*( A->x + B->x ) - sqrt(3.0)/6.0 * ( B->y - A->y );
	fermat_point->y = 0.5*( A->y + B->y ) + sqrt(3.0)/6.0 * ( B->x - A->x );
}

void init_femat_locus_iterator_semi(
	const Point_i* A, const Point_i* B,
	Fermat_locus_iterator* it, int pas
){
	it->pas = pas;
	it->A.x = A->x;
	it->A.y = A->y;
	it->B.x = B->x;
	it->B.y = B->y;
	it->direction.x = B->x - A->x;
	it->direction.y = B->y - A->y;
	couple_i_to_f( A, &(it->A_f) );
	couple_i_to_f( B, &(it->B_f) );
	center_of_fermat_circle( &(it->A_f), &(it->B_f), &(it->fermat_center) );
	it->rayon = distance( &(it->A_f), &(it->fermat_center) );
	it->current.x = A->x;
	it->current.y = A->y;

	it->old.x = -1;
	it->old.y = -1;

	Fermat_locus_iterator copy = *it;
	next_fermat_locus_iterator_semi( &copy );
	Point_i next;
	get_point_from_fermat_locus_iterator( &copy, &next );
	Point_f direction = { (float)( next.x - A->x ), (float)( next.y - A->y) };
	if( vectorial_product( &(it->direction), &direction ) >= 0 ){
		it->old = next;
	}
}

void restart_fermat_locus_iterator_semi( Fermat_locus_iterator* it ){
	init_femat_locus_iterator( &it->A, &it->B, it, it->pas );
}

inline void next_locus_hwd(
	const Point_f* direction, const Point_f* fermat_center,  const Point_f* prev,
	const Point_f* current,
	float* min, Point_f* better, float rayon, const Point_f* old 
){
	float tmp;
	const Point_f delta = {current->x - prev->x,current->y - prev->y };
	if( scalar_product( &delta ,direction ) >= 0 ){
		if( (old->x != current->x) || (old->y != current->y) ){
			tmp = fabs( distance( fermat_center, current ) - rayon );
			if( tmp < *min ){
				*min = tmp;
				*better = *current;
			}
		}
	}
}

inline int next_fermat_locus_iterator_semi( Fermat_locus_iterator* const it ){
	const float pas = it->pas;
	if( 
		( abs( it->current.x - it->B.x ) <= pas ) &&
		( abs( it->current.y - it->B.y ) <= pas )
	){
		return 0;
	}
	Point_f old = { (float) it->old.x,(float) it->old.y };
	float min;
	const float rayon = it->rayon;
	Point_f better;
	Point_f current = { (float) it->current.x, (float) it->current.y };
	Point_f prev = current;
	min = 10*pas + fabs( distance( &(it->fermat_center), &(current) ) - rayon );
	// N
		current.x = prev.x;
		current.y = prev.y+pas;
		next_locus_hwd( 
			&(it->direction), &(it->fermat_center), &prev, 
			&current, &min, &better, rayon, &old 
		);
	// NE
		current.x = prev.x+pas;
		next_locus_hwd( 
			&(it->direction), &(it->fermat_center),  &prev,
			&current, &min, &better, rayon, &old 
		);
	// E
		current.y = prev.y;
		next_locus_hwd( 
			&(it->direction), &(it->fermat_center),  &prev,
			&current, &min, &better, rayon, &old
		);
	// SE
		current.y = prev.y-pas;
		next_locus_hwd( 
			&(it->direction), &(it->fermat_center),  &prev,
			&current, &min, &better, rayon, &old
		);
	// S
		current.x = prev.x;
		next_locus_hwd( 
			&(it->direction), &(it->fermat_center),  &prev,
			&current, &min, &better, rayon, &old 
		);
	// S0
		current.x = prev.x-pas;
		next_locus_hwd( 
			&(it->direction), &(it->fermat_center),  &prev,
			&current, &min, &better, rayon, &old 
		);
	// O
		current.y = prev.y;
		next_locus_hwd( 
			&(it->direction), &(it->fermat_center),  &prev,
			&current, &min, &better, rayon, &old 
		);
	// NO
		current.y = prev.y+pas;
		next_locus_hwd( 
			&(it->direction), &(it->fermat_center),  &prev,
			&current, &min, &better, rayon, &old 
		);
	it->old = it->current;
	couple_f_to_i( &better, &(it->current) );
	return 1;
}

void init_femat_locus_iterator(
	const Point_i* A, const Point_i* B,
	Fermat_locus_iterator* it, int pas
){
	init_femat_locus_iterator_semi( A, B, it, pas);
	it->premiere_moitie = 1;
}
void restart_fermat_locus_iterator( Fermat_locus_iterator* it ){
	Point_i A = it->A;
	Point_i B = it->B;
	if( it->premiere_moitie ){
		init_femat_locus_iterator(	&A, &B, it, it->pas );
	}else{
		init_femat_locus_iterator(	&B, &A, it, it->pas );
	}
}

int next_fermat_locus_iterator( Fermat_locus_iterator* const it ){
	if( next_fermat_locus_iterator_semi( it ) ) return 1;
	if( it->premiere_moitie ){
		const Point_i A = it->A;
		const Point_i B = it->B;
		init_femat_locus_iterator_semi(	&B, &A, it, it->pas );
		it->premiere_moitie = 0;
		return 1;
	}else{
		return 0;
	}
}


void get_point_from_fermat_locus_iterator(
	const Fermat_locus_iterator* it, Point_i* point
){
	point->x = it->current.x;
	point->y = it->current.y;
}

